﻿using System.Drawing;
using System.Windows.Forms;

namespace LocalYuuPic
{
    public static class DarkModeUtility
    {
        // ダークモード時の基本カラー
        private static readonly Color DarkBackColor = Color.FromArgb(32, 32, 32);
        private static readonly Color DarkForeColor = Color.FromArgb(240, 240, 240);
        private static readonly Color DarkControlBackColor = Color.FromArgb(45, 45, 45);
        private static readonly Color DarkControlForeColor = Color.FromArgb(240, 240, 240);
        private static readonly Color DarkButtonBackColor = Color.FromArgb(60, 60, 60);
        private static readonly Color DarkTextBoxBackColor = Color.FromArgb(30, 30, 30);

        /// <summary>
        /// フォームとその子コントロールにダークモードのカラーを適用します
        /// </summary>
        /// <param name="control">対象のコントロールまたはフォーム</param>
        /// <param name="isDarkMode">ダークモードか否か</param>
        public static void ApplyDarkMode(Control control, bool isDarkMode)
        {
            if (control == null) return;

            // コントロールの種類に応じて色を設定
            switch (control)
            {
                case Form form:
                    ApplyFormColors(form, isDarkMode);
                    break;

                case Button button:
                    ApplyButtonColors(button, isDarkMode);
                    break;

                case TextBox textBox:
                    ApplyTextBoxColors(textBox, isDarkMode);
                    break;

                case DataGridView gridView:
                    ApplyDataGridViewColors(gridView, isDarkMode);
                    break;

                case ComboBox comboBox:
                    ApplyComboBoxColors(comboBox, isDarkMode);
                    break;

                case Label label:
                    ApplyLabelColors(label, isDarkMode);
                    break;

                case Panel panel:
                    ApplyPanelColors(panel, isDarkMode);
                    break;

                case GroupBox groupBox:
                    ApplyGroupBoxColors(groupBox, isDarkMode);
                    break;

                case ListView listView:
                    ApplyListViewColors(listView, isDarkMode);
                    break;

                // 他のコントロールタイプはデフォルトの処理
                default:
                    ApplyDefaultControlColors(control, isDarkMode);
                    break;
            }

            // 子コントロールに対して再帰的に処理
            foreach (Control child in control.Controls)
            {
                ApplyDarkMode(child, isDarkMode);
            }
        }

        private static void ApplyFormColors(Form form, bool isDarkMode)
        {
            if (isDarkMode)
            {
                form.BackColor = DarkBackColor;
                form.ForeColor = DarkForeColor;
            }
            else
            {
                form.BackColor = SystemColors.Control;
                form.ForeColor = SystemColors.ControlText;
            }
        }

        private static void ApplyButtonColors(Button button, bool isDarkMode)
        {
            if (isDarkMode)
            {
                button.BackColor = DarkButtonBackColor;
                button.ForeColor = DarkControlForeColor;
                button.FlatStyle = FlatStyle.Flat;
                button.FlatAppearance.BorderColor = Color.FromArgb(100, 100, 100);
            }
            else
            {
                button.BackColor = SystemColors.ButtonHighlight;
                button.ForeColor = SystemColors.ControlText;
                button.FlatStyle = FlatStyle.Standard;
            }
        }

        private static void ApplyTextBoxColors(TextBox textBox, bool isDarkMode)
        {
            if (isDarkMode)
            {
                textBox.BackColor = DarkTextBoxBackColor;
                textBox.ForeColor = DarkControlForeColor;
            }
            else
            {
                textBox.BackColor = SystemColors.Window;
                textBox.ForeColor = SystemColors.WindowText;
            }
        }

        private static void ApplyDataGridViewColors(DataGridView gridView, bool isDarkMode)
        {
            if (isDarkMode)
            {
                gridView.BackgroundColor = DarkControlBackColor;
                gridView.ForeColor = DarkControlForeColor;
                gridView.GridColor = Color.FromArgb(70, 70, 70);
                gridView.DefaultCellStyle.BackColor = DarkControlBackColor;
                gridView.DefaultCellStyle.ForeColor = DarkControlForeColor;
                gridView.DefaultCellStyle.SelectionBackColor = Color.FromArgb(60, 60, 100);
                gridView.ColumnHeadersDefaultCellStyle.BackColor = DarkButtonBackColor;
                gridView.ColumnHeadersDefaultCellStyle.ForeColor = DarkControlForeColor;
                gridView.EnableHeadersVisualStyles = false;
            }
            else
            {
                gridView.BackgroundColor = SystemColors.Window;
                gridView.ForeColor = SystemColors.ControlText;
                gridView.GridColor = SystemColors.Control;
                gridView.DefaultCellStyle.BackColor = SystemColors.Window;
                gridView.DefaultCellStyle.ForeColor = SystemColors.ControlText;
                gridView.DefaultCellStyle.SelectionBackColor = SystemColors.Highlight;
                gridView.ColumnHeadersDefaultCellStyle.BackColor = SystemColors.Control;
                gridView.ColumnHeadersDefaultCellStyle.ForeColor = SystemColors.ControlText;
                gridView.EnableHeadersVisualStyles = true;
            }
        }

        private static void ApplyComboBoxColors(ComboBox comboBox, bool isDarkMode)
        {
            if (isDarkMode)
            {
                comboBox.BackColor = DarkControlBackColor;
                comboBox.ForeColor = DarkControlForeColor;
            }
            else
            {
                comboBox.BackColor = SystemColors.Window;
                comboBox.ForeColor = SystemColors.WindowText;
            }
        }

        private static void ApplyLabelColors(Label label, bool isDarkMode)
        {
            if (isDarkMode)
            {
                label.ForeColor = DarkControlForeColor;
                // 透過の場合はBackColorを設定しない
                if (label.BackColor != Color.Transparent)
                {
                    label.BackColor = DarkControlBackColor;
                }
            }
            else
            {
                label.ForeColor = SystemColors.ControlText;
                if (label.BackColor != Color.Transparent)
                {
                    label.BackColor = SystemColors.Control;
                }
            }
        }

        private static void ApplyPanelColors(Panel panel, bool isDarkMode)
        {
            if (isDarkMode)
            {
                panel.BackColor = DarkControlBackColor;
                panel.ForeColor = DarkControlForeColor;
            }
            else
            {
                panel.BackColor = SystemColors.Control;
                panel.ForeColor = SystemColors.ControlText;
            }
        }

        private static void ApplyGroupBoxColors(GroupBox groupBox, bool isDarkMode)
        {
            if (isDarkMode)
            {
                groupBox.BackColor = DarkControlBackColor;
                groupBox.ForeColor = DarkControlForeColor;
            }
            else
            {
                groupBox.BackColor = SystemColors.Control;
                groupBox.ForeColor = SystemColors.ControlText;
            }
        }

        private static void ApplyListViewColors(ListView listView, bool isDarkMode)
        {
            if (isDarkMode)
            {
                listView.BackColor = DarkControlBackColor;
                listView.ForeColor = DarkControlForeColor;
            }
            else
            {
                listView.BackColor = SystemColors.Window;
                listView.ForeColor = SystemColors.ControlText;
            }
        }

        private static void ApplyDefaultControlColors(Control control, bool isDarkMode)
        {
            if (isDarkMode)
            {
                control.BackColor = DarkControlBackColor;
                control.ForeColor = DarkControlForeColor;
            }
            else
            {
                control.BackColor = SystemColors.Control;
                control.ForeColor = SystemColors.ControlText;
            }
        }
    }
}
