using System;
using System.Drawing;
using System.IO;
using WIC;
using ImageMagick;
using ImageMagick.Factories;
using LocalYuuPic;
using System.Diagnostics;
using System.Windows.Forms;
using Dark.Net;

public class ImageProcessing : IDisposable
{
    private Bitmap _bitmap;

    public enum ImageFormat
    {
        Png,
        Jpeg,
        WebPLossless,
        WebPLossy,
        HeicLossless,
        HeicLossy
    }

    public int Width => _bitmap.Width;
    public int Height => _bitmap.Height;
    public ImageFormat Format;

    private static ImagePreviewForm previewForm;
    private const int PreviewWidth = 300;
    private const int PreviewHeight = 200;
    private static Form ownerForm;

    public ImageProcessing(string filePath, bool isMaintainTransparency)
    {
        using (var image = new MagickImage(filePath))
        {
            // 画像フォーマットの判定と設定
            switch (image.Format)
            {
                case MagickFormat.Png:
                    Format = ImageFormat.Png;
                    break;
                case MagickFormat.Jpeg:
                    Format = ImageFormat.Jpeg;
                    break;
                case MagickFormat.WebP:
                    Format = (image.Quality == 100) ? ImageFormat.WebPLossless : ImageFormat.WebPLossy;
                    break;
                case MagickFormat.Heic:
                    Format = (image.Quality == 100) ? ImageFormat.HeicLossless : ImageFormat.HeicLossy;
                    break;
                default:
                    throw new NotSupportedException($"未対応のフォーマット: {image.Format}");
            }

            if (!isMaintainTransparency)
            {
                image.Alpha(AlphaOption.Remove);
                image.BackgroundColor = new MagickColor("white");
            }
            _bitmap = new Bitmap(image.ToBitmap());
        }
    }

    public ImageProcessing(Bitmap bmp)
    {
        _bitmap = new Bitmap(bmp);
    }

    public void Save(Stream outputStream, ImageFormat format, uint quality = 0)
    {
        switch (format)
        {
            case ImageFormat.Png:
                quality = 95; //zlib=9,adaptive
                break;
            case ImageFormat.WebPLossless:
            case ImageFormat.HeicLossless:
                quality = 100;
                break;
            case ImageFormat.Jpeg:
                if (quality == 0)
                {
                    quality = 80;
                }
                break;
            case ImageFormat.WebPLossy:
            case ImageFormat.HeicLossy:
                if (quality == 0)
                {
                    quality = 75;
                }
                break;
        }

        switch (format)
        {
            case ImageFormat.Png:
            case ImageFormat.Jpeg:
            case ImageFormat.WebPLossless:
            case ImageFormat.WebPLossy:
                SaveWithMagick(outputStream, format, quality);
                break;
            case ImageFormat.HeicLossless:
            case ImageFormat.HeicLossy:
                SaveHeic(outputStream, format == ImageFormat.HeicLossless, quality);
                break;
            default:
                throw new ArgumentException("Unsupported image format", nameof(format));
        }
    }

    private void SaveWithMagick(Stream outputStream, ImageFormat format, uint quality)
    {
        var m = new MagickFactory();
        using (var image = new MagickImage(m.Image.Create(_bitmap)))
        {
            image.Quality = quality;

            MagickFormat magickFormat;
            switch (format)
            {
                case ImageFormat.Png:
                    magickFormat = MagickFormat.Png;
                    image.Settings.SetDefine(MagickFormat.Png, "compression-level", "9");
                    image.Settings.SetDefine(MagickFormat.Png, "compression-strategy", "4");
                    image.SetAttribute("png:exclude-chunks", "tEXt,cHRM,pHYs");
                    break;
                case ImageFormat.Jpeg:
                    magickFormat = MagickFormat.Jpeg;
                    image.Settings.SetDefine(MagickFormat.Jpeg, "sampling-factor", "4:2:0");
                    image.Alpha(AlphaOption.Remove); // JPEG doesn't support transparency
                    break;
                case ImageFormat.WebPLossless:
                case ImageFormat.WebPLossy:
                    magickFormat = MagickFormat.WebP;
                    image.Settings.SetDefine(MagickFormat.WebP, "lossless", format == ImageFormat.WebPLossless ? "true" : "false");

                    if (format == ImageFormat.WebPLossy)
                    {
                        image.Settings.SetDefine(MagickFormat.WebP, "method", "4"); // バランスの取れた圧縮方法
                    }
                    break;
                default:
                    throw new ArgumentException("Unsupported format for Magick.NET", nameof(format));
            }

            image.Write(outputStream, magickFormat);
        }
    }

    private void SaveHeic(Stream outputStream, bool isLossless, uint quality)
    {
        if (_bitmap == null)
            throw new InvalidOperationException("Bitmap is not initialized");

        if (quality > 100)
            throw new ArgumentOutOfRangeException(nameof(quality), "Quality must be between 0 and 100");

        var factory = new WICImagingFactory();

        // HEIFエンコーダーの作成
        var encoder = factory.CreateEncoder(ContainerFormat.Heif);

        encoder.Initialize(outputStream.AsCOMStream(), WICBitmapEncoderCacheOption.WICBitmapEncoderNoCache);

        // フレームの作成
        var frame = encoder.CreateNewFrame(out IPropertyBag2 props);

        frame.Initialize(props);

        // ビットマップデータの変換と書き込み
        using (var memStream = new MemoryStream())
        {
            _bitmap.Save(memStream, System.Drawing.Imaging.ImageFormat.Bmp);
            memStream.Position = 0;

            var decoder = factory.CreateDecoderFromStream(memStream, WICDecodeOptions.WICDecodeMetadataCacheOnDemand);

            var sourceBitmap = decoder.GetFrame(0);
                
            frame.WriteSource(sourceBitmap, null);
        }

        frame.Commit();

        try
        {
            encoder.Commit();
        }
        catch (Exception ex)
        {
            Debug.WriteLine($"Error SaveHeic: {ex.Message}\n\nStacktrace:\n{ex.StackTrace}");
        }
    }

    public void Resize(int width, int height)
    {
        if ( _bitmap.Width == width && _bitmap.Height == height ) { return; }

        var destImage = new Bitmap(width, height);
        using (var graphics = Graphics.FromImage(destImage))
        {
            graphics.CompositingQuality = System.Drawing.Drawing2D.CompositingQuality.HighQuality;
            graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
            graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
            graphics.DrawImage(_bitmap, 0, 0, width, height);
        }
        Dispose();
        _bitmap = destImage;
    }

    public double CalculateResizeRatio(int targetSizeKB, ImageFormat format)
    {
        double minRatio = 0.1;
        double maxRatio = 1.0;
        double currentRatio = maxRatio;  // 最初は最大比率から開始
        double bestRatio = minRatio;  // 目標サイズ以下の最大の比率を記録
        int maxIterations = 8;  // イテレーション数
        double tolerance = 0.05;  // 許容誤差

        for (int i = 0; i < maxIterations; i++)
        {
            using (var sizeStream = new SizeCalculatingStream())
            {
                int newWidth = (int)(_bitmap.Width * currentRatio);
                int newHeight = (int)(_bitmap.Height * currentRatio);
                using (var resized = FromBitmap(_bitmap))
                {
                    resized.Resize(newWidth, newHeight);
                    resized.Save(sizeStream, format);
                }

                long currentSize = sizeStream.Length / 1024;

                if (currentSize <= targetSizeKB)
                {
                    // 目標サイズ以下なら、bestRatioを更新
                    bestRatio = currentRatio;
                    minRatio = currentRatio;
                    currentRatio = (currentRatio + maxRatio) / 2;
                }
                else
                {
                    // サイズが大きい場合、比率を下げる
                    maxRatio = currentRatio;
                    currentRatio = (minRatio + currentRatio) / 2;
                }

                // 探索範囲が十分に狭まった場合に終了
                if (maxRatio - minRatio < tolerance)
                {
                    break;
                }
            }
        }

        return bestRatio;
    }

    public void Dispose()
    {
        _bitmap?.Dispose();
    }

    public static ImageProcessing FromBitmap(Bitmap bitmap)
    {
        return new ImageProcessing(bitmap);
    }

    public ImageFormat GetImageFormat(string fileType)
    {
        switch (fileType.ToLower().Replace(" ", "").Replace("\t", ""))
        {
            case "png":
                return ImageFormat.Png;
            case "jpeg":
                return ImageFormat.Jpeg;
            case "webp(lossless)":
                return ImageFormat.WebPLossless;
            case "webp(lossy)":
                return ImageFormat.WebPLossy;
            case "heic(lossless)":
                return ImageFormat.HeicLossless;
            case "heic(lossy)":
                return ImageFormat.HeicLossy;
            default: //auto
                return Format;
        }
    }

    public string GetOutputExtension(string fileType)
    {
        switch (fileType.ToLower().Replace(" ","").Replace("\t", ""))
        {
            case "png":
                return ".png";
            case "jpeg":
                return ".jpg";
            case "webp(lossless)":
            case "webp(lossy)":
                return ".webp";
            case "heic(lossless)":
            case "heic(lossy)":
                return ".heic";
        }

        //defalt (auto)
        switch (Format)
        {
            case ImageFormat.Png:
                return ".png";
            case ImageFormat.Jpeg:
                return ".jpg";
            case ImageFormat.WebPLossless:
            case ImageFormat.WebPLossy:
                return ".webp";
            case ImageFormat.HeicLossless:
            case ImageFormat.HeicLossy:
                return ".heic";
            default:
                throw new NotImplementedException();
        }
    }

    public static void InitializePreview(Form owner)
    {
        ownerForm = owner;
    }

    public static void ShowPreview(string filePath, Point screenLocation)
    {
        if (!File.Exists(filePath)) return;

        var fileInfo = new FileInfo(filePath);
        if (fileInfo.Length == 0) return;

        try
        {
            // プレビューフォームの遅延初期化
            if (previewForm == null || previewForm.IsDisposed)
            {
                previewForm = new ImagePreviewForm(ownerForm);
                DarkNet.Instance.SetWindowThemeForms(previewForm, Theme.Auto);
            }
            previewForm.ApplyTheme(DarkNet.Instance.EffectiveCurrentProcessThemeIsDark);

            using (var processor = new ImageProcessing(filePath, true))
            {
                // プレビュー用の縮小画像を作成
                double scale = Math.Min(
                    (double)PreviewWidth / processor.Width,
                    (double)PreviewHeight / processor.Height
                );
                int width = (int)(processor.Width * scale);
                int height = (int)(processor.Height * scale);

                var previewImage = new Bitmap(width, height);
                using (var g = Graphics.FromImage(previewImage))
                {
                    g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
                    g.DrawImage(processor._bitmap, 0, 0, width, height);
                }

                // ツールチップテキストを生成
                string tooltipText = $"{Path.GetFileName(filePath)}\n" +
                                   $"形式: {GetFormatDisplayName(processor.Format)}\n" +
                                   $"サイズ: {FormatFileSize(fileInfo.Length)}\n" +
                                   $"寸法: {processor.Width} x {processor.Height}";

                // プレビューを表示
                previewForm.ShowPreview(tooltipText, previewImage, screenLocation);
            }
        }
        catch (Exception ex)
        {
            Debug.WriteLine($"プレビュー生成エラー: {ex.Message}");
            HidePreview();
        }
    }

    public static void HidePreview()
    {
        previewForm?.Hide();
    }

    public static void CleanupPreview()
    {
        if (previewForm != null && !previewForm.IsDisposed)
        {
            previewForm.Dispose();
            previewForm = null;
        }
    }

    // ファイルサイズのフォーマット用ヘルパーメソッド
    public static string FormatFileSize(long bytes)
    {
        string[] suffixes = { "B", "KB", "MB", "GB", "TB" };
        int counter = 0;
        decimal number = bytes;
        while (Math.Round(number / 1024) >= 1)
        {
            number = number / 1024;
            counter++;
        }
        return string.Format("{0:n1}{1}", number, suffixes[counter]);
    }

    // フォーマットの表示名を取得するメソッドを追加
    public static string GetFormatDisplayName(ImageFormat format)
    {
        switch (format)
        {
            case ImageFormat.Png:
                return "PNG";
            case ImageFormat.Jpeg:
                return "JPEG";
            case ImageFormat.WebPLossless:
                return "WebP (ロスレス)";
            case ImageFormat.WebPLossy:
                return "WebP (非可逆)";
            case ImageFormat.HeicLossless:
                return "HEIC (ロスレス)";
            case ImageFormat.HeicLossy:
                return "HEIC (非可逆)";
            default:
                return "不明な形式";
        }
    }
}