﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Dark.Net;

namespace LocalYuuPic
{
    public partial class MainForm : Form
    {
        private string[] validExtensions = { ".png", ".jpg", ".jpeg", ".webp", ".heic" };
        private enum ResizeMode
        {
            ByPercentage,
            ByPixel,
            ByFileSize
        }

        private struct ConversionSettings
        {
            public List<string> FilePaths;
            public string OutputDirectory;
            public string FileType;
            public string FilePostfix;
            public bool FilePrefixNum;
            public ResizeMode ResizeMode;
            public int ResizePercent;
            public int ResizePixel;
            public int TargetFileSize;
            public bool MaintainTransparency;
            public bool WarnFileOverwrite;
            public bool ClearAfterComplete;
        }

        private class MessageBoxParameters
        {
            public string Text { get; set; }
            public string Caption { get; set; }
            public MessageBoxButtons Buttons { get; set; }
            public MessageBoxIcon Icon { get; set; }

            public MessageBoxParameters(
                string text,
                string caption = "",
                MessageBoxButtons buttons = MessageBoxButtons.OK,
                MessageBoxIcon icon = MessageBoxIcon.None)
            {
                Text = text;
                Caption = caption;
                Buttons = buttons;
                Icon = icon;
            }
        }

        //=============================================================================================================

        private ContextMenuStrip listViewContextMenu;
        private bool isLoadingSettings = false;

        private int insertionIndex = -1;
        private bool insertAfterLastItem = false;
        private Rectangle insertionLine = Rectangle.Empty;

        private string lastPreviewPath = null;

        //==============================================================================================================

        public MainForm()
        {
            if (Properties.Settings.Default.UpgradeRequired)
            {
                Properties.Settings.Default.Upgrade();
                Properties.Settings.Default.UpgradeRequired = false;
                Properties.Settings.Default.Save();
            }

            InitializeComponent();

            checkableGroupResize = ReplaceGroupBox(checkableGroupResize);

            LoadSettings();

            fileList.Columns[0].Width = -2;

            SetupContextMenu();

            DarkNet.Instance.EffectiveCurrentProcessThemeIsDarkChanged += (_, isDarkTheme) => ApplyTheme(isDarkTheme);
            ApplyTheme(DarkNet.Instance.EffectiveCurrentProcessThemeIsDark);

            previewTimer.Interval = 500; // 500ms

            ImageProcessing.InitializePreview(this);
        }

        public void ApplyTheme(bool isDarkTheme)
        {
            DarkModeUtility.ApplyDarkMode(this, isDarkTheme);
        }

        private CheckableGroupBox ReplaceGroupBox(GroupBox oldGroupBox)
        {
            // 新しいCheckableGroupBoxを作成
            var newGroupBox = new CheckableGroupBox
            {
                Name = oldGroupBox.Name,
                Text = oldGroupBox.Text,
                Location = oldGroupBox.Location,
                Size = oldGroupBox.Size,
                Anchor = oldGroupBox.Anchor,
                Dock = oldGroupBox.Dock,
                Tag = oldGroupBox.Tag,
                Checked = true  // デフォルトの状態を設定
            };

            // 元のGroupBox内のコントロールを新しいGroupBoxに移動
            while (oldGroupBox.Controls.Count > 0)
            {
                var control = oldGroupBox.Controls[0];
                oldGroupBox.Controls.Remove(control);
                newGroupBox.Controls.Add(control);
            }

            // フォームのControlsコレクションで置き換え
            var index = Controls.IndexOf(oldGroupBox);
            Controls.RemoveAt(index);
            Controls.Add(newGroupBox);
            newGroupBox.BringToFront();  // Z順序を維持

            // イベントハンドラの再接続（必要な場合）
            if (oldGroupBox.Tag is string eventName && !string.IsNullOrEmpty(eventName))
            {
                // イベントハンドラ名がTagに保存されている場合の例
                var method = this.GetType().GetMethod(eventName,
                    System.Reflection.BindingFlags.NonPublic |
                    System.Reflection.BindingFlags.Instance);
                if (method != null)
                {
                    newGroupBox.CheckedChanged += (EventHandler)Delegate.CreateDelegate(
                        typeof(EventHandler), this, method);
                }
            }

            return newGroupBox;
        }

        private void LoadSettings()
        {
            isLoadingSettings = true;

            try
            {
                // 出力先の設定 - outputDirの中身で判断
                if (string.IsNullOrEmpty(Properties.Settings.Default.OutputDirectory))
                {
                    radioOutputSame.Checked = true;
                }
                else
                {
                    radioOutputDir.Checked = true;
                    outputDir.Text = Properties.Settings.Default.OutputDirectory;
                }

                var cb = (CheckableGroupBox)checkableGroupResize;
                cb.Checked = Properties.Settings.Default.ResizeEnable;

                // ファイル名後置
                filePostfix.Text = Properties.Settings.Default.FilePostfix;

                filePrefixNum.Checked = Properties.Settings.Default.FilePrefixNum;

                // ファイル形式の設定
                comboFileType.SelectedIndex = Properties.Settings.Default.FileType;
                checkBoxMaintainTransparency.Checked = Properties.Settings.Default.MaintainTransparency;

                checkClearAfterComplete.Checked = Properties.Settings.Default.ClearAfterComplete;
                checkWarnFileOverwrite.Checked = Properties.Settings.Default.WarnFileOverwrite;

                // リサイズの数値設定を選択状態に関係なく読み込み
                sizePercent.Text = Properties.Settings.Default.ResizePercent;
                sizePX.Value = Properties.Settings.Default.ResizePixel;
                targetFileSize.Value = Properties.Settings.Default.TargetFileSize;

                // リサイズモードの選択状態を設定
                switch (Properties.Settings.Default.ResizeMode)
                {
                    case "Percent":
                        radioResizePercent.Checked = true;
                        break;
                    case "Pixel":
                        radioResizePX.Checked = true;
                        break;
                    case "FileSize":
                        radioResizeFilesize.Checked = true;
                        break;
                }

                // 設定が存在しない場合のフォールバック
                if (comboFileType.SelectedIndex < 0) comboFileType.SelectedIndex = 0;
                if (string.IsNullOrEmpty(sizePercent.Text)) sizePercent.SelectedIndex = 0;
            }
            finally
            {
                isLoadingSettings = false;
            }
        }

        private void SaveSettings()
        {
            // 別フォルダに出力が選択されている場合のみ出力先を保存
            Properties.Settings.Default.OutputDirectory = radioOutputDir.Checked ? outputDir.Text : string.Empty;

            // ファイル名後置
            Properties.Settings.Default.FilePostfix = filePostfix.Text;

            Properties.Settings.Default.FilePrefixNum = filePrefixNum.Checked;

            // ファイル形式の設定
            Properties.Settings.Default.FileType = comboFileType.SelectedIndex;
            Properties.Settings.Default.MaintainTransparency = checkBoxMaintainTransparency.Checked;

            Properties.Settings.Default.ClearAfterComplete = checkClearAfterComplete.Checked;
            Properties.Settings.Default.WarnFileOverwrite = checkWarnFileOverwrite.Checked;

            // リサイズ設定 - モード選択
            Properties.Settings.Default.ResizeMode = radioResizePercent.Checked ? "Percent" :
                                                   radioResizePX.Checked ? "Pixel" : "FileSize";

            // リサイズ設定 - 選択状態に関係なくすべての値を保存
            Properties.Settings.Default.ResizePercent = sizePercent.Text;
            Properties.Settings.Default.ResizePixel = sizePX.Value;
            Properties.Settings.Default.TargetFileSize = targetFileSize.Value;

            var cb = (CheckableGroupBox)checkableGroupResize;
            Properties.Settings.Default.ResizeEnable = cb.Checked;

            Properties.Settings.Default.Save();
        }

        private void SetupContextMenu()
        {
            listViewContextMenu = new ContextMenuStrip();
            listViewContextMenu.Items.Add("名前でソート (昇順)", null, (sender, e) => SortListView("name", SortOrder.Ascending));
            listViewContextMenu.Items.Add("名前でソート (降順)", null, (sender, e) => SortListView("name", SortOrder.Descending));
            listViewContextMenu.Items.Add("サイズでソート (昇順)", null, (sender, e) => SortListView("size", SortOrder.Ascending));
            listViewContextMenu.Items.Add("サイズでソート (降順)", null, (sender, e) => SortListView("size", SortOrder.Descending));
            listViewContextMenu.Items.Add("更新日時でソート (昇順)", null, (sender, e) => SortListView("date", SortOrder.Ascending));
            listViewContextMenu.Items.Add("更新日時でソート (降順)", null, (sender, e) => SortListView("date", SortOrder.Descending));

            fileList.ContextMenuStrip = listViewContextMenu;
        }

        private void SortListView(string sortBy, SortOrder sortOrder)
        {
            List<ListViewItem> items = new List<ListViewItem>();
            foreach (ListViewItem item in fileList.Items)
            {
                items.Add(item);
            }

            switch (sortBy)
            {
                case "name":
                    items = sortOrder == SortOrder.Ascending
                        ? items.OrderBy(item => item.Text).ToList()
                        : items.OrderByDescending(item => item.Text).ToList();
                    break;
                case "size":
                    items = sortOrder == SortOrder.Ascending
                        ? items.OrderBy(item => new FileInfo(item.Tag.ToString()).Length).ToList()
                        : items.OrderByDescending(item => new FileInfo(item.Tag.ToString()).Length).ToList();
                    break;
                case "date":
                    items = sortOrder == SortOrder.Ascending
                        ? items.OrderBy(item => new FileInfo(item.Tag.ToString()).LastWriteTime).ToList()
                        : items.OrderByDescending(item => new FileInfo(item.Tag.ToString()).LastWriteTime).ToList();
                    break;
            }

            fileList.BeginUpdate();
            fileList.Items.Clear();
            fileList.Items.AddRange(items.ToArray());
            fileList.EndUpdate();
        }

        private void MainForm_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
                e.Effect = DragDropEffects.Copy;
        }

        private void MainForm_DragDrop(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);
                ProcessDroppedPaths(files, fileList.Items.Count); // ファイルをリストの最後に追加
            }
        }

        private void ProcessDroppedPaths(string[] paths, int insertIndex)
        {
            foreach (string path in paths)
            {
                if (File.Exists(path))
                {
                    ProcessFile(path, insertIndex++);
                }
                else if (Directory.Exists(path))
                {
                    ProcessDirectory(path, ref insertIndex);
                }
            }
        }

        private void ProcessFile(string filePath, int insertIndex)
        {
            if (validExtensions.Contains(Path.GetExtension(filePath).ToLower()))
            {
                // ファイルパスが既にリストに存在するか確認
                bool fileExists = fileList.Items.Cast<ListViewItem>().Any(item => (string)item.Tag == filePath);

                if (!fileExists)
                {
                    Icon fileIcon = Icon.ExtractAssociatedIcon(filePath);
                    imageList.Images.Add(filePath, fileIcon.ToBitmap());

                    ListViewItem item = new ListViewItem(Path.GetFileName(filePath), imageList.Images.Count - 1);
                    item.Tag = filePath;
                    if (insertIndex > fileList.Items.Count)
                    {
                        fileList.Items.Add(item);
                    }
                    else
                    {
                        fileList.Items.Insert(insertIndex, item);
                    }
                }
                else
                {
                    System.Media.SystemSounds.Exclamation.Play();
                }
            }
        }

        private void ProcessDirectory(string directoryPath, ref int insertIndex)
        {
            var files = Directory.EnumerateFiles(directoryPath, "*.*", SearchOption.AllDirectories)
                                 .Where(file => validExtensions.Contains(Path.GetExtension(file).ToLower()));

            foreach (string file in files)
            {
                ProcessFile(file, insertIndex++);
            }
        }

        private void clearFileList_Click(object sender, EventArgs e)
        {
            fileList.Items.Clear();
            imageList.Images.Clear();
        }

        private void clearSelectedList_Click(object sender, EventArgs e)
        {
            // 選択されたアイテムを配列に格納し、インデックスの大きい順にソート
            var selectedItems = fileList.SelectedItems.Cast<ListViewItem>()
                .OrderByDescending(item => item.Index)
                .ToArray();

            // 選択されているアイテムがない場合は何もしない
            if (!selectedItems.Any()) return;

            fileList.BeginUpdate();
            try
            {
                foreach (ListViewItem item in selectedItems)
                {
                    // ImageListから対応する画像を削除
                    if (item.ImageIndex >= 0)
                    {
                        imageList.Images.RemoveAt(item.ImageIndex);
                        // 削除した後のインデックスを更新
                        foreach (ListViewItem remainingItem in fileList.Items)
                        {
                            if (remainingItem.ImageIndex > item.ImageIndex)
                            {
                                remainingItem.ImageIndex--;
                            }
                        }
                    }

                    // ListViewからアイテムを削除
                    fileList.Items.Remove(item);
                }
            }
            finally
            {
                fileList.EndUpdate();
            }
        }

        private void radioOutputDir_CheckedChanged(object sender, EventArgs e)
        {
            if (isLoadingSettings) return;

            if (radioOutputDir.Checked)
            {
                using (FolderBrowserDialog folderBrowserDialog = new FolderBrowserDialog())
                {
                    folderBrowserDialog.Description = "出力先ディレクトリを選択してください";
                    folderBrowserDialog.ShowNewFolderButton = true;

                    if (folderBrowserDialog.ShowDialog() == DialogResult.OK)
                    {
                        outputDir.Text = folderBrowserDialog.SelectedPath;
                    }
                    else
                    {
                        radioOutputSame.Checked = true;
                    }
                }
            }
            else {
                outputDir.Text = "";
            }
        }

        private void buttonExecute_Click(object sender, EventArgs e)
        {
            if (fileList.Items.Count == 0)
            {
                ShowMessageBoxSafe("変換する画像ファイルを追加してください。", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            var cb = (CheckableGroupBox)checkableGroupResize;
            var noResize = !cb.Checked;

            ResizeMode resizeMode = radioResizePercent.Checked ? ResizeMode.ByPercentage :
                                    radioResizePX.Checked ? ResizeMode.ByPixel :
                                    ResizeMode.ByFileSize;

            List<string> filePaths = new List<string>();
            foreach (ListViewItem item in fileList.Items)
            {
                filePaths.Add(item.Tag.ToString());
            }

            ConversionSettings settings = new ConversionSettings
            {
                FilePaths = filePaths,
                OutputDirectory = radioOutputDir.Checked ? outputDir.Text : string.Empty,
                FileType = comboFileType.SelectedItem.ToString().Split(new string[] { " - " }, 2, StringSplitOptions.None)[0],
                FilePostfix = filePostfix.Text,
                FilePrefixNum = filePrefixNum.Checked,
                ResizeMode = noResize ? ResizeMode.ByPercentage : resizeMode,
                ResizePercent = noResize ? 100 : int.Parse(sizePercent.Text.ToString().Split(' ')[0]),
                ResizePixel = int.Parse(sizePX.Text.ToString()),
                TargetFileSize = (int)targetFileSize.Value,
                MaintainTransparency = checkBoxMaintainTransparency.Checked,
                WarnFileOverwrite = checkWarnFileOverwrite.Checked,
                ClearAfterComplete = checkClearAfterComplete.Checked,
            };

            buttonExecute.Enabled = false;
            buttonCancel.Enabled = true;
            progressBar.Value = 0;

            UpdateOutputResult("実行中");

            imageWorker.RunWorkerAsync(settings);
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            if (imageWorker.IsBusy)
            {
                imageWorker.CancelAsync();
            }
        }

        private DialogResult ShowMessageBoxSafeInvoke(MessageBoxParameters parameters)
        {
            var oldTxt = outputResult.Text;

            outputResult.Text = "**問い合わせ中**";

            var r = MessageBox.Show(
                this,
                parameters.Text,
                parameters.Caption,
                parameters.Buttons,
                parameters.Icon
            );

            outputResult.Text = oldTxt;

            return r;
        }
        private DialogResult ShowMessageBoxSafe(MessageBoxParameters parameters)
        {
            if (InvokeRequired)
            {
                // Invokeを使用して結果を取得
                return (DialogResult)Invoke(new Func<MessageBoxParameters, DialogResult>(ShowMessageBoxSafeInvoke), parameters);
            }

            return ShowMessageBoxSafeInvoke(parameters);
        }

        // オーバーロードバージョン
        private DialogResult ShowMessageBoxSafe(
            string text,
            string caption = "",
            MessageBoxButtons buttons = MessageBoxButtons.OK,
            MessageBoxIcon icon = MessageBoxIcon.None)
        {
            var parameters = new MessageBoxParameters(text, caption, buttons, icon);
            return ShowMessageBoxSafe(parameters);
        }

        private void UpdateOutputResultInvoke(string text)
        {
            outputResult.Text = text;
        }

        private void UpdateOutputResult(string text)
        {
            if (outputResult.InvokeRequired)
            {
                // メインスレッドにディスパッチ
                outputResult.Invoke(new Action<string>(UpdateOutputResultInvoke), text);
                return;
            }

            UpdateOutputResultInvoke(text);
        }

        private void imageWorker_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            try
            {
                imageWorker_DoWork_Real(sender, e);
            }
            catch (Exception ex)
            {
                ShowMessageBoxSafe($"エラーが発生しました: {ex.Message}", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void imageWorker_DoWork_Real(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            ConversionSettings settings = (ConversionSettings)e.Argument;
            BackgroundWorker worker = sender as BackgroundWorker;

            int totalFiles = settings.FilePaths.Count;
            int processedFiles = 0;

            var outputPaths = new List<string>();

            for (int i = 0; i < totalFiles; i++)
            {
                string opath;

                if ( ! DetermineOutputPath(i, totalFiles, settings.FilePaths[i], settings, out opath) )
                {
                    e.Cancel = true;
                    return;
                }

                outputPaths.Add(opath);
            }

            // Calculate the degree of parallelism
            int degreeOfParallelism = Environment.ProcessorCount > 1 ? Environment.ProcessorCount - 1 : 1;
            if (degreeOfParallelism > 4) { degreeOfParallelism = 4; }

            // Create a partitioner to split the work
            var partitioner = Partitioner.Create(0, totalFiles);

            // Use ParallelOptions to set the degree of parallelism
            ParallelOptions parallelOptions = new ParallelOptions
            {
                MaxDegreeOfParallelism = degreeOfParallelism
            };

            try
            {
                Parallel.ForEach(partitioner, parallelOptions, (range, loopState) =>
                {
                    for (int i = range.Item1; i < range.Item2; i++)
                    {
                        if (worker.CancellationPending)
                        {
                            loopState.Stop();
                            e.Cancel = true;
                            return;
                        }

                        string filePath = settings.FilePaths[i];
                        string outputPath = outputPaths[i];

                        try
                        {
                            ConvertImage(filePath, outputPath, settings);
                        }
                        catch (Exception ex)
                        {
                            Debug.WriteLine($"Error converting {filePath}: {ex.Message}\n\nStacktrace:\n{ex.StackTrace}");
                        }

                        // Increment the processed files count and report progress
                        int processed = Interlocked.Increment(ref processedFiles);
                        worker.ReportProgress(processed * 100 / totalFiles);
                    }
                });
            }
            catch (OperationCanceledException)
            {
                e.Cancel = true;
            }
        }

        private void ConvertImage(string inputPath, string outputPath, ConversionSettings settings)
        {
            using (var imageProcessor = new ImageProcessing(inputPath, settings.MaintainTransparency))
            {
                var outputFormat = imageProcessor.GetImageFormat(settings.FileType);

                outputPath += imageProcessor.GetOutputExtension(settings.FileType);

                int newWidth = imageProcessor.Width;
                int newHeight = imageProcessor.Height;

                switch (settings.ResizeMode)
                {
                    case ResizeMode.ByPercentage:
                        double percentageRatio = settings.ResizePercent / 100.0;
                        if (percentageRatio < 1.0)
                        {
                            newWidth = (int)Math.Round(imageProcessor.Width * percentageRatio);
                            newHeight = (int)Math.Round(imageProcessor.Height * percentageRatio);
                        }
                        break;

                    case ResizeMode.ByFileSize:
                        double fileSizeRatio = imageProcessor.CalculateResizeRatio(settings.TargetFileSize * 1024, outputFormat);
                        if (fileSizeRatio < 1.0)
                        {
                            newWidth = (int)Math.Round(imageProcessor.Width * fileSizeRatio);
                            newHeight = (int)Math.Round(imageProcessor.Height * fileSizeRatio);
                        }
                        break;

                    case ResizeMode.ByPixel:
                        int longSide = Math.Max(imageProcessor.Width, imageProcessor.Height);
                        if (settings.ResizePixel < longSide)
                        {
                            double resizeRatio = (double)settings.ResizePixel / longSide;
                            newWidth = (int)Math.Round(imageProcessor.Width * resizeRatio);
                            newHeight = (int)Math.Round(imageProcessor.Height * resizeRatio);
                        }
                        break;
                }

                imageProcessor.Resize(newWidth, newHeight);
                using (var outputStream = new FileStream(outputPath, FileMode.Create))
                {
                    imageProcessor.Save(outputStream, outputFormat);
                }
            }
        }

        public static List<string> FindExistingFiles(string baseFilePath)
        {
            if (string.IsNullOrEmpty(baseFilePath))
            {
                return new List<string>();
            }

            try
            {
                // ファイルパスをディレクトリとファイル名に分割
                string directory = Path.GetDirectoryName(baseFilePath) ?? ".";
                string fileNameWithoutExtension = Path.GetFileNameWithoutExtension(baseFilePath);

                // ディレクトリが存在しない場合は空リストを返す
                if (!Directory.Exists(directory))
                {
                    return new List<string>();
                }

                // 指定されたファイル名で始まるファイルを検索
                return Directory.GetFiles(directory, $"{fileNameWithoutExtension}.*")
                              .ToList();
            }
            catch (Exception ex)
            {
                // 例外が発生した場合はログに記録するなどの処理を追加可能
                Console.WriteLine($"Error searching for files: {ex.Message}");
                return new List<string>();
            }
        }

        private bool DetermineOutputPath(int curNum, int totalNum, string inputPath, ConversionSettings settings, out string outputPath)
        {
            //拡張子を除くファイル名を生成する
            string directory = string.IsNullOrEmpty(settings.OutputDirectory)
                ? Path.GetDirectoryName(inputPath)
                : settings.OutputDirectory;
            string fileName = Path.GetFileNameWithoutExtension(inputPath);

            // FilePrefixNumが有効な場合、番号を付加
            if (settings.FilePrefixNum)
            {
                // totalNumの桁数を計算
                int digits = totalNum.ToString().Length;
                // curNumを指定桁数で0埋め
                string numberPrefix = curNum.ToString($"D{digits}") + "_";
                fileName = numberPrefix + fileName;
            }

            // 元のフォルダに出力する場合は ファイル名 を付加
            if (string.IsNullOrEmpty(settings.OutputDirectory))
            {
                fileName += settings.FilePostfix;
            }

            outputPath = Path.Combine(directory, fileName);

            var list = FindExistingFiles(outputPath);

            if (list.Count > 0)
            {
                if (settings.WarnFileOverwrite)
                {
                    if (ShowMessageBoxSafe($"以下のファイルを上書きしますか？\n{outputPath}", "ファイル上書き警告", MessageBoxButtons.OKCancel, MessageBoxIcon.Exclamation) == DialogResult.Cancel)
                    {
                        return false;
                    }
                }
                foreach (var file in list)
                {
                    File.Delete(file);
                }
            }

            return true;

            // ファイル名の衝突を回避
            /*int counter = 1;
            while (File.Exists(outputPath))
            {
                outputPath = Path.Combine(directory, $"{fileName}_{counter}");
                counter++;
            }*/
        }


        private void imageWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            progressBar.Value = e.ProgressPercentage;
        }

        private void imageWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            buttonExecute.Enabled = true;
            buttonCancel.Enabled = false;

            if (e.Cancelled)
            {
                UpdateOutputResult("キャンセル");
            }
            else
            {
                UpdateOutputResult("変換完了");

                if ( checkClearAfterComplete.Checked ) {
                    clearFileList_Click(sender, e);
                }
            }
        }


        private void fileList_ItemDrag(object sender, ItemDragEventArgs e)
        {
            fileList.DoDragDrop(fileList.SelectedItems, DragDropEffects.Move);
        }

        private void fileList_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(typeof(ListView.SelectedListViewItemCollection)))
            {
                e.Effect = DragDropEffects.Move;
            }
            else if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                e.Effect = DragDropEffects.Copy;
            }
        }

        private void fileList_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(typeof(ListView.SelectedListViewItemCollection))
                || e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                Point cp = fileList.PointToClient(new Point(e.X, e.Y));
                int newInsertionIndex = GetInsertionIndex(cp);

                if (newInsertionIndex != insertionIndex || insertAfterLastItem)
                {
                    insertionIndex = newInsertionIndex;
                    fileList.Invalidate();
                }

                e.Effect = DragDropEffects.Move;
            }
        }
        private int GetInsertionIndex(Point clientPoint)
        {
            ListViewItem item = fileList.GetItemAt(clientPoint.X, clientPoint.Y);
            insertAfterLastItem = false;

            if (item == null)
            {
                insertAfterLastItem = true;
                return fileList.Items.Count;
            }

            Rectangle itemBounds = item.Bounds;
            if (clientPoint.Y < itemBounds.Top + (itemBounds.Height / 2))
            {
                return item.Index;
            }
            else
            {
                if (item.Index == fileList.Items.Count - 1)
                {
                    insertAfterLastItem = true;
                }
                return item.Index + 1;
            }
        }

        private void fileList_DragDrop(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(typeof(ListView.SelectedListViewItemCollection)))
            {
                ListView.SelectedListViewItemCollection draggedItems =
                    (ListView.SelectedListViewItemCollection)e.Data.GetData(typeof(ListView.SelectedListViewItemCollection));

                List<ListViewItem> itemsToMove = new List<ListViewItem>();
                foreach (ListViewItem item in draggedItems)
                {
                    itemsToMove.Add(item);
                }

                int targetIndex = insertionIndex;
                if (insertAfterLastItem)
                {
                    targetIndex = fileList.Items.Count;
                }

                // ドラッグされたアイテムを元のリストから削除する前に、挿入位置を調整
                int removedCount = 0;
                foreach (ListViewItem item in itemsToMove)
                {
                    int oldIndex = fileList.Items.IndexOf(item);
                    if (oldIndex < targetIndex)
                    {
                        removedCount++;
                    }
                    fileList.Items.Remove(item);
                }
                targetIndex -= removedCount;

                // 調整後の位置にアイテムを挿入
                foreach (ListViewItem item in itemsToMove)
                {
                    fileList.Items.Insert(targetIndex, item);
                    targetIndex++;
                }
            }
            else if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);
                int targetIndex = insertAfterLastItem ? fileList.Items.Count : insertionIndex;
                ProcessDroppedPaths(files, targetIndex);
            }

            // ドラッグ＆ドロップ操作が完了したらインサーションマークをクリア
            insertionIndex = -1;
            insertAfterLastItem = false;
            fileList.Invalidate();
        }

        private void fileList_DrawItem(object sender, DrawListViewItemEventArgs e)
        {
            e.DrawDefault = true;

            if (insertionIndex >= 0)
            {
                using (Pen pen = new Pen(Color.Red, 2))
                {
                    if (insertAfterLastItem && e.ItemIndex == fileList.Items.Count - 1)
                    {
                        // 最後のアイテムの下に線を引く
                        e.Graphics.DrawLine(pen, e.Bounds.Left, e.Bounds.Bottom, e.Bounds.Right, e.Bounds.Bottom);
                    }
                    else if (e.ItemIndex == insertionIndex)
                    {
                        // アイテムの上に線を引く
                        e.Graphics.DrawLine(pen, e.Bounds.Left, e.Bounds.Top, e.Bounds.Right, e.Bounds.Top);
                    }
                }
            }
        }

        private void fileList_DrawColumnHeader(object sender, DrawListViewColumnHeaderEventArgs e)
        {
            e.DrawDefault = true;
        }

        private void labelHelp_Click(object sender, EventArgs e)
        {
            var message = @"このウィンドウ内の任意の場所に、変換したいファイルをドラッグ＆ドロップしてください。
フォルダをドロップすると、フォルダ内のすべてのファイルが追加されます。
ファイルリストはCtrl+クリックやShift+クリックで複数選択できます。
ファイルリストの上でマウスをしばらく止めると画像をプレビューできます。

EXIF情報は自動的に消去され、変換後のファイルには残りません。

PNG: イラスト向き。無圧縮で画質劣化なし。
JPEG: 写真向き。圧縮で画質が劣化。
WebP（Lossless）: イラスト向き。圧縮しても画質劣化なし。
WebP（Lossy）: 写真向き。圧縮で画質が一部劣化。
HEIC（Lossy）: 写真向き。高圧縮。圧縮で画質が一部劣化。

古い技術←　PNG/JPEG　WebP　HEIC　→新しい技術

新しい技術のフォーマットを使うと、圧縮率が向上しますが、互換性は低下します。
";
            ShowMessageBoxSafe(
                 message,
                 "詳しい解説",
                 MessageBoxButtons.OK,
                 MessageBoxIcon.Information
             );
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            SaveSettings();

            previewTimer.Dispose();
            ImageProcessing.CleanupPreview();
        }

        private void fileList_MouseMove(object sender, MouseEventArgs e)
        {
            ListViewItem item = fileList.GetItemAt(e.X, e.Y);

            if (item != null)
            {
                string filePath = item.Tag.ToString();
                if (filePath == lastPreviewPath) return;

                lastPreviewPath = filePath;
                previewTimer.Stop();
                previewTimer.Tag = new { FilePath = filePath, Location = fileList.PointToScreen(e.Location) };
                previewTimer.Start();
            }
            else
            {
                lastPreviewPath = null;
                previewTimer.Stop();
                ImageProcessing.HidePreview();
            }
        }

        private void fileList_MouseLeave(object sender, EventArgs e)
        {
            lastPreviewPath = null;
            previewTimer.Stop();
            ImageProcessing.HidePreview();
        }

        private void previewTimer_Tick(object sender, EventArgs e)
        {
            previewTimer.Stop();
            var data = (dynamic)previewTimer.Tag;
            ImageProcessing.ShowPreview(data.FilePath, data.Location);
        }
    }
}
